/* RCSid: $Id: mgf_mesg.h,v 3.1 2011/02/18 00:40:25 greg Exp $ */
/*
 * Various messages, English version
 */

/* The following goes into the mg_err array, defined in parser.h */

#define MG_ERRLIST	{"No error",\
			"Unknown entity",\
			"Wrong number of arguments",\
			"Wrong argument type",\
			"Illegal argument value",\
			"Undefined reference",\
			"Cannot open input file",\
			"Error in included file",\
			"Out of memory",\
			"Seek failure",\
			"Illegal material specification",\
			"Input line too long",\
			"Unmatched context close",\
			}
